# Coef
import numpy as np
from sklearn.linear_model import LinearRegression

def get_linear_coef(series):
    lr = LinearRegression()
    x = np.arange(1, len(series) + 1).reshape(-1, 1)
    y = series.values.reshape(-1, 1)
    lr.fit(x, y)
    coef = lr.coef_[0][0]
    return coef


def get_normalized_linear_coef(series):
    normalized_series = series / series.mean()
    return get_linear_coef(normalized_series)


import pandas as pd

from .utils import get_linear_coef,get_normalized_linear_coef
from .utils import filter_date
from .utils import calc_profit

class MultiOhlcvAnalyser:
    def __init__(self, multi_ohlcv) -> None:
        self.ohlcv = multi_ohlcv.copy()

    @staticmethod
    def _get_groupby_code(ohlcv, start, end):
        groupby_ohlcv = filter_date(ohlcv, start, end).groupby("code")
        return groupby_ohlcv

    def info(self, start=None, end=None):
        ohlcv = filter_date(self.ohlcv, start, end)
        groupby_code = self._get_groupby_code(ohlcv, start, end)
        ps = groupby_code['close'].apply(lambda x: calc_profit(x))
        info_dict = {
            "total_stock_code": len(set(ohlcv["code"])),
            "start_date": ohlcv.index.min(),
            "end_date": ohlcv.index.max(),
            "market_average_profit": (ps.mean().round(2)) * 100,
            "increased_stock_pct": round(len(ps[ps > 0]) / len(ps), 2) * 100,
            "decreased_stock_pct": round(len(ps[ps <= 0]) / len(ps), 2) * 100,
        }
        info_df = pd.DataFrame.from_dict(
            info_dict, orient="index", columns=["value"]
        )
        return info_df

    def coef(self, arg, start=None, end=None):
        groupby_code = self._get_groupby_code(self.ohlcv, start, end)
        coef_series = groupby_code.apply(
            lambda x: get_linear_coef(x[arg])
        ).rename(f"{arg}_coef")
        return coef_series

    def normalized_coef(self, arg, start=None, end=None):
        groupby_code = self._get_groupby_code(self.ohlcv, start, end)
        normalized_coef_series = groupby_code.apply(
            lambda x: get_normalized_linear_coef(x[arg])
        ).rename(f"{arg}_normalized_coef")
        return normalized_coef_series

    def oc_variance(self, start, end):
        """
        (open - close) variance
        """
        ohlcv = self.ohlcv.copy()
        ohlcv["oc_change"] = (ohlcv["open"] - ohlcv["close"]) / ohlcv["open"]
        groupby_code = self._get_groupby_code(ohlcv, start, end)
        oc_variance_series = groupby_code.apply(
            lambda x: x["oc_change"].var()
        ).rename("oc_variance")
        return oc_variance_series

    def hl_variance(self, start, end):
        """
        (high - low) variance
        """
        ohlcv = self.ohlcv.copy()
        ohlcv["hl_change"] = (ohlcv["high"] - ohlcv["low"]) / ohlcv["low"]
        groupby_code = self._get_groupby_code(ohlcv, start, end)
        hl_variance_series = groupby_code.apply(
            lambda x: x["hl_change"].var()
        ).rename("hl_variance")
        return hl_variance_series
    
    def profit(self, start, end):
        ohlcv = self.ohlcv.copy()
        groupby_code = self._get_groupby_code(ohlcv, start, end)
        profit_series = groupby_code.apply(
            lambda x: calc_profit(x)
        ).rename("profit")
        return profit_series


import numpy as np
import pandas as pd


class ProfitAnalyser:
    @staticmethod
    def calc_pct(value):
        return round(value * 100, 3)

    def get_start_end_profit(self, price_series):
        """
        (start,end) Profit Calculator
        """
        buying_price = price_series.iloc[0]
        selling_price = price_series.iloc[-1]
        profit = (selling_price - buying_price) / buying_price
        return self.calc_pct(profit)

    def get_start_max_profit(self, price_series):
        """
        (start, max) Max Profit Calculator
        """
        buying_price = price_series.iloc[0]
        max_price = price_series.max()
        max_profit = (max_price - buying_price) / buying_price
        return self.calc_pct(max_profit)

    def get_start_min_profit(self, price_series):
        """
        (start, min) Min Profit Calculator
        """
        buying_price = price_series.iloc[0]
        min_price = price_series.min()
        min_profit = (min_price - buying_price) / buying_price
        return self.calc_pct(min_profit)

    def get_statistical_price_series(self, high_series, low_series):
        """
        (high, low) based Noramlized distribution assumption
        """
        mean_series = ((high_series + low_series) / 2).rename("mean")
        var_series = ((high_series - low_series) / 4).rename("var")

        def get_value_based_on_normal_dist_assumption(mean, var):
            return np.random.normal(mean, var, 1)[0]

        statistical_price_series = pd.concat(
            [mean_series, var_series], axis=1
        ).apply(
            lambda x: get_value_based_on_normal_dist_assumption(
                x["mean"], x["var"]
            ),
            axis=1,
        )
        return statistical_price_series



import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression

lr = LinearRegression()


class CoefficientAnalyser:
    @staticmethod
    def get_coefficient(series):
        x = np.arange(1, len(series) + 1).reshape(-1, 1)
        y = series.values.reshape(-1, 1)
        lr.fit(x, y)
        coefficient = lr.coef_[0][0]
        return coefficient

    @staticmethod
    def get_coefficient_score(series):
        x = np.arange(1, len(series) + 1).reshape(-1, 1)
        y = series.values.reshape(-1, 1)
        lr.fit(x, y)
        score = lr.score(x, y)
        return score

    @staticmethod
    def get_normalized_coefficient(series):
        normalized_series = series / series.mean()
        normalized_coefficient = CoefficientAnalyser.get_coefficient(
            normalized_series
        )
        return normalized_coefficient

    @staticmethod
    def get_statistical_price_series(high_series, low_series):
        """
        (high, low) based Noramlized distribution assumption
        """
        mean_series = ((high_series + low_series) / 2).rename("mean")
        var_series = ((high_series - low_series) / 4).rename("var")

        def get_value_based_on_normal_dist_assumption(mean, var):
            return np.random.normal(mean, var, 1)[0]

        statistical_price_series = pd.concat(
            [mean_series, var_series], axis=1
        ).apply(
            lambda x: get_value_based_on_normal_dist_assumption(
                x["mean"], x["var"]
            ),
            axis=1,
        )
        return statistical_price_series



import numpy as np
import pandas as pd


class PriceAnalyser:
    @staticmethod
    def get_normalized_series(series, n=1):
        noramlized_series = (series / series.mean()) * n
        return round(noramlized_series).astype(int)

    @staticmethod
    def get_general_prices(price_series, volume_series):
        normalized_volume_series = PriceAnalyser.get_normalized_series(
            volume_series, 10
        )
        prices = np.concatenate(
            (
                price_series.apply(lambda x: [x]) * normalized_volume_series
            ).values
        )
        return prices

    @staticmethod
    def get_statistical_prices(high_series, low_series, volume_series):
        normalized_volume_series = PriceAnalyser.get_normalized_series(
            volume_series, 10
        )
        mean_series = ((high_series + low_series) / 2).rename("mean")
        var_series = ((high_series - low_series) / 4).rename("var")

        def get_value_based_on_normal_dist_assumption(mean, var, volume):
            return np.random.normal(mean, var, int(volume))

        statistical_prices = np.concatenate(
            pd.concat(
                [
                    mean_series,
                    var_series,
                    normalized_volume_series,
                ],
                axis=1,
            )
            .apply(
                lambda x: get_value_based_on_normal_dist_assumption(
                    x["mean"], x["var"], x["volume"]
                ),
                axis=1,
            )
            .values
        )

        return statistical_prices

    def get_price_rank(self, prices, price):
        """
        price upper n%
        """
        sorted_purchases = np.sort(np.append(prices, price))
        rank = np.where(sorted_purchases == price)[0][0] + 1
        percentile = round(
            ((len(sorted_purchases) - rank) / len(sorted_purchases)) * 100, 2
        )
        return percentile
